import { Ref, ComputedRef } from 'vue-demi';
import { IHistoryOptions } from 'stateshot';
import { ResizeObserver } from '@juggle/resize-observer';
import { adjustHue, darken, desaturate, getContrast, getLuminance, getScale, guard, hasBadContrast, hsla, lighten, mix, opacify, parseToHsla, readableColor, readableColorIsBlack, rgba, saturate, transparentize, parseToRgba, ColorError, toHex, toRgba, toHsla } from 'color2k';

declare type Fn = () => void;

declare type RefTyped<T> = T | Ref<T>;
declare type RefElement = Element | Ref<Element | undefined>;

/**
 * Size of a specific box.
 *
 * https://drafts.csswg.org/resize-observer-1/#resizeobserversize
 */
interface ResizeObserverSize {
    readonly inlineSize: number;
    readonly blockSize: number;
}
/**
 * Used to observe different box sizes.
 *
 * https://drafts.csswg.org/resize-observer-1/#enumdef-resizeobserverboxoptions
 */
declare enum ResizeObserverBoxOptions {
    BORDER_BOX = "border-box",
    CONTENT_BOX = "content-box",
    DEVICE_PIXEL_CONTENT_BOX = "device-pixel-content-box"
}
/**
 * Options to be given to the resize observer,
 * when observing a new element.
 *
 * https://drafts.csswg.org/resize-observer-1/#dictdef-resizeobserveroptions
 */
interface ResizeObserverOptions {
    box?: 'content-box' | 'border-box' | 'device-pixel-content-box' | ResizeObserverBoxOptions;
}
/**
 * https://drafts.csswg.org/resize-observer-1/#resize-observer-entry-interface
 */
interface ResizeObserverEntry {
    readonly target: Element;
    readonly contentRect: DOMRectReadOnly;
    readonly borderBoxSize?: ReadonlyArray<ResizeObserverSize>;
    readonly contentBoxSize?: ReadonlyArray<ResizeObserverSize>;
    readonly devicePixelContentBoxSize?: ReadonlyArray<ResizeObserverSize>;
}
/**
 * https://drafts.csswg.org/resize-observer-1/#resize-observer-callback
 */
declare type ResizeObserverCallback = (entries: ResizeObserverEntry[], observer: ResizeObserver) => void;

interface WindowSize {
    width: Ref<number>;
    height: Ref<number>;
}
interface ElementSize extends WindowSize {
}
interface EventOptions {
    useCapture?: boolean;
    useThrottle?: boolean | null;
    delay?: number;
}

interface ReactiveColor {
    hexColor: ComputedRef<string>;
    hslaColor: ComputedRef<string>;
    rgbaColor: ComputedRef<string>;
    readableColor: ComputedRef<string>;
    invertColor: ComputedRef<string>;
    luminance: ComputedRef<number>;
    lightenColor: ComputedRef<string>;
    darkenColor: ComputedRef<string>;
    tintColor: ComputedRef<string>;
    shadeColor: ComputedRef<string>;
    desaturateColor: ComputedRef<string>;
    saturateColor: ComputedRef<string>;
    opacifyColor: ComputedRef<string>;
    transparentizeColor: ComputedRef<string>;
}

declare const useCssVars: (el: RefTyped<HTMLElement | null>, variable: string, defaultValue: string) => Ref<string>;

/**
 * enhanced `color2k` with new utilities like `tint` 、 `shade` 、 `invert`.
 */
declare const usePalette: {
    adjustHue: typeof adjustHue;
    darken: typeof darken;
    desaturate: typeof desaturate;
    getContrast: typeof getContrast;
    getLuminance: typeof getLuminance;
    getScale: typeof getScale;
    guard: typeof guard;
    hasBadContrast: typeof hasBadContrast;
    hsla: typeof hsla;
    lighten: typeof lighten;
    mix: typeof mix;
    opacify: typeof opacify;
    parseToHsla: typeof parseToHsla;
    readableColor: typeof readableColor;
    readableColorIsBlack: typeof readableColorIsBlack;
    rgba: typeof rgba;
    saturate: typeof saturate;
    transparentize: typeof transparentize;
    parseToRgba: typeof parseToRgba;
    ColorError: typeof ColorError;
    toHex: typeof toHex;
    toRgba: typeof toRgba;
    toHsla: typeof toHsla;
    /**
     * Tints a color by mixing it with white.
     * @param color
     * @param amount given as a decimal between 0 and 1
     */
    tint: (color: string, amount: number) => string;
    /**
     * Shades a color by mixing it with black.
     * @param color
     * @param amount given as a decimal between 0 and 1
     */
    shade: (color: string, amount: number) => string;
    /**
     * Inverts the red, green and blue values of a color.
     * @param color
     */
    invert: (color: string) => string;
};
/**
 * Reactive color parsing and manipulation built on top of color2k.
 * @param color
 * @param amount
 */
declare const useColor: (color: Ref<string>, amount?: Ref<number> | undefined) => ReactiveColor;

declare const useDebouncedRef: (value: any, delay?: number) => Ref;

declare const useElementSize: (target: Element, options: ResizeObserverOptions) => ElementSize;

declare function useEventListener<E extends keyof WindowEventMap>(target: Window, event: E, listener: (this: Window, ev: WindowEventMap[E]) => any, options?: boolean | AddEventListenerOptions): Fn;
declare function useEventListener<E extends keyof DocumentEventMap>(target: Document, event: E, listener: (this: Document, ev: DocumentEventMap[E]) => any, options?: boolean | AddEventListenerOptions): Fn;
declare function useEventListener<E extends keyof DocumentEventMap>(target: RefElement | Ref<any>, event: E, listener: (this: Element, ev: DocumentEventMap[E]) => any, options?: boolean | AddEventListenerOptions): Fn;
declare function useEventListener(target: RefTyped<EventTarget>, event: string, listener: EventListener, options?: boolean | AddEventListenerOptions): Fn;

interface IFullScreen {
    isFullscreen: Ref<boolean>;
    toggleFullscreen: Fn;
}
/**
 * Reactive Fullscreen API build on top of screenfull.js
 * @param {HTMLElement} target
 * @return {boolean} isFullscreen
 * @return {function} toggleFullscreen
 */
declare const useFullscreen: (target: HTMLElement) => IFullScreen;

declare const useInstance: () => any;

/**
 * Call onMounted() if it's inside a component lifecycle, if not, run just call the function
 * @param fn
 * @param sync if set to false, it will run in the nextTick() of Vue
 */
declare function tryOnMounted(fn: Fn, sync?: boolean): void;
/**
 * Call onUnmounted() if it's inside a component lifecycle, if not, do nothing
 * @param fn
 */
declare function tryOnUnmounted(fn: Fn): void;

/**
 * Register `MouseMove` event on mounted, and remove event automatically on unmounted.
 */
declare const useMouse: () => {
    x: Ref<number>;
    y: Ref<number>;
    remove: () => void;
};

/**
 * `usePrecision` provide reactive methods to perform addition, subtraction, multiplication & division operations precisely.
 */
declare const usePrecision: () => {
    strip: (num: RefTyped<string | number>, precision?: RefTyped<number | undefined>) => ComputedRef<number>;
    plus: (num1: RefTyped<string | number>, num2: RefTyped<string | number>, ...others: RefTyped<string | number>[]) => ComputedRef<number>;
    minus: (num1: RefTyped<string | number>, num2: RefTyped<string | number>, ...others: RefTyped<string | number>[]) => ComputedRef<number>;
    times: (num1: RefTyped<string | number>, num2: RefTyped<string | number>, ...others: RefTyped<string | number>[]) => ComputedRef<number>;
    divide: (num1: RefTyped<string | number>, num2: RefTyped<string | number>, ...others: RefTyped<string | number>[]) => ComputedRef<number>;
    round: (num: RefTyped<string | number>, ratio: RefTyped<number>) => ComputedRef<number>;
    digitLength: (num: RefTyped<string | number>) => ComputedRef<number>;
    float2Fixed: (num: RefTyped<string | number>) => ComputedRef<number>;
};

declare function useReactiveState(getter: () => Required<object>): [Required<object>, VoidFunction];

/**
 * Reactive Resize API, When window resize, do someting.
 * @param {function} fn
 * @param {object} options
 * @return {function} dispatchResize
 */
declare const useResizeEvent: (fn: Fn, options?: EventOptions) => {
    dispatchResize: () => void;
};

/**
 * A map of keybinding strings to event handlers.
 */
interface KeyBindingMap {
    [keybinding: string]: (event: KeyboardEvent) => void;
}
interface KeymasterEvent {
    key: string;
    method: KeyHandler;
    mods: number[];
    scope: string;
    shortcut: string;
}
interface KeyHandler {
    (keyboardEvent: KeyboardEvent, keymasterEvent: KeymasterEvent): void;
}
/**
 * Bind shortcut onMounted & unbind shortcut automatically onUnmounted in setup hooks
 * or
 * Unbind shortcut before bind shortcut without setup hooks
 * @param {Object} keymap
 * @param {Object} vm
 * @return {function} toggle
 */
declare const useShortcut: (keymap: KeyBindingMap, vm: any) => {
    shift: boolean;
    alt: boolean;
    option: boolean;
    ctrl: boolean;
    control: boolean;
    command: boolean;
    setScope(scopeName: string): void;
    getScope(): string;
    deleteScope(scopeName: string): void;
    noConflict(): void;
    unbind(key: string): void;
    unbind(key: string, scopeName: string): void;
    isPressed(keyCode: number): boolean;
    getPressedKeyCodes(): number[];
    filter(event: FilterEvent): void;
    bind: (seed: string, func: Function) => void;
};

/**
 * 💾 A state snapshot composable utilities built on top of stateshot.js.
 * @param defaultValue
 * @param options checkout stateshot [options](https://github.com/gaoding-inc/stateshot#history)
 */
declare const useStateshot: (defaultValue: RefTyped<any>, options?: IHistoryOptions) => {
    undo: () => void;
    redo: () => void;
    reset: () => void;
    jump: (step: number) => any;
    snapshot: (state: any) => void;
    historyState: Ref<any>;
    historyLength: Ref<number>;
    hasRedo: Ref<boolean>;
    hasUndo: Ref<boolean>;
    undoCount: Ref<number>;
    redoCount: Ref<number>;
};

/**
 * It immediately detects when an element resizes and provides accurate sizing information back to the handler.
 * @param target
 * @param callback
 * @param options
 */
declare const useResizeObserver: (target: Element, callback: ResizeObserverCallback, options: ResizeObserverOptions) => {
    stop: () => void;
};

/**
 * Reactive window size
 * @param {object} options
 * @return {Ref<number>} WindowSize
 */
declare const useWindowSize: (options: EventOptions) => WindowSize;

declare const isArray: (arg: any) => arg is any[];
declare const isBoolean: (val: any) => val is boolean;
declare const isClient: boolean;
declare const isElement: (val: any) => val is Element;
declare const isUndefined: (val: any) => boolean;

declare const NOOP: () => void;

declare function unwrap<T>(o: RefTyped<T>): T;
declare function wrap(o: RefElement): Ref<Element>;
declare function wrap<T>(o: RefTyped<T>): Ref<T>;
declare type ReactiveFn<T> = T extends (...args: infer A) => infer R ? (...args: {
    [K in keyof A]: RefTyped<A[K]>;
}) => ComputedRef<R> : never;
declare function reactiveFn<T extends Function>(fn: T): ReactiveFn<T>;

export { IFullScreen, KeyBindingMap, KeyHandler, KeymasterEvent, NOOP, ReactiveFn, isArray, isBoolean, isClient, isElement, isUndefined, reactiveFn, tryOnMounted, tryOnUnmounted, unwrap, useColor, useCssVars, useDebouncedRef, useElementSize, useEventListener, useFullscreen, useInstance, useMouse, usePalette, usePrecision, useReactiveState, useResizeEvent, useResizeObserver, useShortcut, useStateshot, useWindowSize, wrap };
