/*!
 * @vueblocks/vue-use-core v0.2.6
 * (c) 2021 xiaoluoboding
 * @license MIT
 */
import { ref, watch, isRef, computed, unref, customRef, getCurrentInstance, onMounted, nextTick, onUnmounted, isVue3, reactive, toRefs } from 'vue-demi';
import * as color2k from 'color2k';
import { mix, parseToRgba, rgba, toHex, toHsla, toRgba, readableColor, getLuminance, lighten, darken, desaturate, saturate, opacify, transparentize } from 'color2k';
import debounce from 'lodash.debounce';
import { ResizeObserver } from '@juggle/resize-observer';
import screenfull from 'screenfull';
import { strip, plus, minus, times, divide, round, digitLength, float2Fixed } from 'number-precision';
import throttle from 'lodash.throttle';
import keymaster from 'keymaster';
import { History } from 'stateshot';
import cloneDeep from 'lodash.clonedeep';

const cssVariableRegex = /--[\S]*/g;
const useCssVars = (el, variable, defaultValue) => {
    if (!variable || !variable.match(cssVariableRegex)) {
        console.warn(`[useCssVars] css variable is not valid: ${variable}`);
    }
    const propertyValue = ref('');
    const target = ref(el || window.document.documentElement);
    watch(target, (el) => {
        if (window && el) {
            const propValue = window.getComputedStyle(el).getPropertyValue(variable) || defaultValue || '';
            propertyValue.value = propValue;
        }
    }, { immediate: true });
    watch(propertyValue, (newVal) => {
        if (target.value && target.value.style) {
            target.value.style.setProperty(variable, newVal);
        }
    });
    return propertyValue;
};

const isArray = Array.isArray;
const isBoolean = (val) => typeof val === 'boolean';
const isClient = ![typeof window, typeof document].includes('undefined');
const isElement = (val) => typeof val === 'object' && !!val.tagName;
const isUndefined = (val) => val === undefined;

const NOOP = () => { };

function unwrap(o) {
    return isRef(o) ? o.value : o;
}
function wrap(o) {
    return isRef(o) ? o : ref(o); // NOTE in v3 this is not necessary
}
function reactiveFn(fn) {
    return function (...args) {
        return computed(() => fn.apply(this, args.map(v => unref(v))));
    };
}

/**
 * enhanced `color2k` with new utilities like `tint` 、 `shade` 、 `invert`.
 */
const usePalette = {
    /**
     * Tints a color by mixing it with white.
     * @param color
     * @param amount given as a decimal between 0 and 1
     */
    tint: (color, amount) => mix('#FFF', color, amount),
    /**
     * Shades a color by mixing it with black.
     * @param color
     * @param amount given as a decimal between 0 and 1
     */
    shade: (color, amount) => mix('#000', color, amount),
    /**
     * Inverts the red, green and blue values of a color.
     * @param color
     */
    invert: (color) => {
        if (color === 'transparent')
            return color;
        const [r, g, b, a] = parseToRgba(color);
        return rgba(255 - r, 255 - g, 255 - b, a);
    },
    ...color2k
};
/**
 * Reactive color parsing and manipulation built on top of color2k.
 * @param color
 * @param amount
 */
const useColor = (color, amount) => {
    const { tint, shade, invert } = usePalette;
    const colorState = {
        hexColor: computed(() => ''),
        hslaColor: computed(() => ''),
        rgbaColor: computed(() => ''),
        readableColor: computed(() => ''),
        invertColor: computed(() => ''),
        luminance: computed(() => 0),
        lightenColor: computed(() => ''),
        darkenColor: computed(() => ''),
        tintColor: computed(() => ''),
        shadeColor: computed(() => ''),
        desaturateColor: computed(() => ''),
        saturateColor: computed(() => ''),
        opacifyColor: computed(() => ''),
        transparentizeColor: computed(() => '')
    };
    // accpet color argument
    if (isRef(color) && unref(color)) {
        colorState.hexColor = computed(() => toHex(unref(color)));
        colorState.hslaColor = computed(() => toHsla(unref(color)));
        colorState.rgbaColor = computed(() => toRgba(unref(color)));
        colorState.readableColor = computed(() => readableColor(unref(color)));
        colorState.invertColor = computed(() => invert(unref(color)));
        colorState.luminance = computed(() => getLuminance(unref(color)));
    }
    // accept amount argument
    if (isRef(amount) && !isUndefined(unref(amount))) {
        colorState.lightenColor = computed(() => lighten(unref(color), unref(amount)));
        colorState.darkenColor = computed(() => darken(unref(color), unref(amount)));
        colorState.tintColor = computed(() => tint(unref(color), unref(amount)));
        colorState.shadeColor = computed(() => shade(unref(color), unref(amount)));
        colorState.desaturateColor = computed(() => desaturate(unref(color), unref(amount)));
        colorState.saturateColor = computed(() => saturate(unref(color), unref(amount)));
        colorState.opacifyColor = computed(() => opacify(unref(color), unref(amount)));
        colorState.transparentizeColor = computed(() => transparentize(unref(color), unref(amount)));
    }
    return colorState;
};

const useDebouncedRef = (value, delay = 200) => {
    return customRef((track, trigger) => {
        return {
            get() {
                track();
                return value;
            },
            set: debounce((newValue) => {
                value = newValue;
                trigger();
            }, delay)
        };
    });
};

/**
 * It immediately detects when an element resizes and provides accurate sizing information back to the handler.
 * @param target
 * @param callback
 * @param options
 */
const useResizeObserver = (target, callback, options) => {
    let ro;
    const targetRef = ref(target);
    const isSupported = window && 'ResizeObserver' in window;
    const stopWatch = watch(targetRef, (newVal) => {
        if (isSupported && window && newVal) {
            ro = new ResizeObserver(callback);
            ro.observe(newVal, options);
        }
    }, { immediate: true });
    const stop = () => {
        ro && ro.disconnect();
        ro = undefined;
        stopWatch();
    };
    return {
        stop
    };
};

const useElementSize = (target, options) => {
    const width = ref(isElement(target) ? target.clientWidth : 0);
    const height = ref(isElement(target) ? target.clientHeight : 0);
    useResizeObserver(target, ([entry]) => {
        width.value = entry.contentRect.width;
        height.value = entry.contentRect.height;
    }, options);
    return { width, height };
};

function useEventListener(target, event, listener, options) {
    let remove = NOOP;
    if (target) {
        const element = ref(target);
        const removeEventListener = (e) => e.removeEventListener(event, listener);
        const addEventListener = (e) => e.addEventListener(event, listener, options);
        const removeWatch = watch(element, (n, _, cleanUp) => {
            if (n) {
                addEventListener(n);
                cleanUp(() => removeEventListener(n));
            }
        }, { immediate: true });
        remove = () => {
            removeEventListener(element.value);
            removeWatch();
        };
    }
    return remove;
}
// export { useEventListener }

/**
 * Call onMounted() if it's inside a component lifecycle, if not, run just call the function
 * @param fn
 * @param sync if set to false, it will run in the nextTick() of Vue
 */
function tryOnMounted(fn, sync = true) {
    if (getCurrentInstance()) {
        onMounted(fn);
    }
    else if (sync) {
        fn();
    }
    else {
        nextTick(fn);
    }
}
/**
 * Call onUnmounted() if it's inside a component lifecycle, if not, do nothing
 * @param fn
 */
function tryOnUnmounted(fn) {
    if (getCurrentInstance())
        onUnmounted(fn);
}

/**
 * Reactive Fullscreen API build on top of screenfull.js
 * @param {HTMLElement} target
 * @return {boolean} isFullscreen
 * @return {function} toggleFullscreen
 */
const useFullscreen = (target) => {
    const targetRef = ref(target || document.querySelector('html'));
    const isFullscreen = ref(false);
    const toggleFullscreen = () => {
        screenfull.isEnabled && screenfull.toggle(targetRef.value);
    };
    tryOnMounted(() => {
        if (screenfull.isEnabled) {
            screenfull.on('change', () => {
                isFullscreen.value = screenfull.isFullscreen;
            });
        }
    });
    return {
        isFullscreen,
        toggleFullscreen
    };
};

const useInstance = () => {
    const instance = getCurrentInstance();
    if (!instance) {
        console.error('`getCurrentInstance` only works during setup or Lifecycle Hooks');
        return null;
    }
    return isVue3 ? instance.ctx : instance.proxy;
};

/**
 * Register `MouseMove` event on mounted, and remove event automatically on unmounted.
 */
const useMouse = () => {
    let remove = NOOP;
    const x = ref(0);
    const y = ref(0);
    // const touchHandler = (event: TouchEvent) => {
    //   if (event.touches.length > 0) {
    //     x.value = event.touches[0].clientX
    //     y.value = event.touches[0].clientY
    //   }
    // }
    if (window) {
        const mouseHandler = (event) => {
            x.value = event.pageX;
            y.value = event.pageY;
        };
        remove = useEventListener(window, 'mousemove', mouseHandler, { passive: true });
    }
    return {
        x,
        y,
        remove
    };
};

/**
 * `usePrecision` provide reactive methods to perform addition, subtraction, multiplication & division operations precisely.
 */
const usePrecision = () => {
    return {
        strip: reactiveFn(strip),
        plus: reactiveFn(plus),
        minus: reactiveFn(minus),
        times: reactiveFn(times),
        divide: reactiveFn(divide),
        round: reactiveFn(round),
        digitLength: reactiveFn(digitLength),
        float2Fixed: reactiveFn(float2Fixed)
    };
};

function useReactiveState(getter) {
    const state = reactive(getter());
    const update = () => {
        Object.assign(state, getter());
    };
    return [state, update];
}

/**
 * Reactive Resize API, When window resize, do someting.
 * @param {function} fn
 * @param {object} options
 * @return {function} dispatchResize
 */
const useResizeEvent = (fn, options = { useCapture: false, useThrottle: null, delay: 500 }) => {
    const { useCapture = false, useThrottle = null, delay = 500 } = options;
    const dispatchResize = () => {
        const e = document.createEvent('Event');
        e.initEvent('resize', true, true);
        window.dispatchEvent(e);
    };
    const waitFn = isBoolean(useThrottle)
        ? useThrottle
            ? throttle(fn, delay)
            : debounce(fn, delay)
        : fn;
    tryOnMounted(() => {
        window.addEventListener('resize', waitFn, useCapture);
    });
    tryOnUnmounted(() => {
        window.removeEventListener('resize', waitFn, useCapture);
    });
    return {
        dispatchResize
    };
};

const bindKeyHandler = (fn) => {
    return (...args) => {
        fn(...args);
        return false;
    };
};
/**
 * Given shortcut a new bind function
 */
const shortcut = {
    bind: (seed, func) => keymaster(seed, bindKeyHandler(func)),
    ...keymaster
};
/**
 * Bind shortcut onMounted & unbind shortcut automatically onUnmounted in setup hooks
 * or
 * Unbind shortcut before bind shortcut without setup hooks
 * @param {Object} keymap
 * @param {Object} vm
 * @return {function} toggle
 */
const useShortcut = (keymap, vm) => {
    let sync = true;
    if (!vm) {
        vm = getCurrentInstance();
    }
    else {
        sync = false;
        unbindKeys();
    }
    function bindKeys() {
        Object.keys(keymap).forEach(key => {
            shortcut.bind(key, keymap[key]);
        });
    }
    function unbindKeys() {
        Object.keys(keymap).forEach(key => {
            shortcut.unbind(key);
        });
    }
    tryOnMounted(() => bindKeys(), sync);
    tryOnUnmounted(unbindKeys);
    return shortcut;
};

/**
 * 💾 A state snapshot composable utilities built on top of stateshot.js.
 * @param defaultValue
 * @param options checkout stateshot [options](https://github.com/gaoding-inc/stateshot#history)
 */
const useStateshot = (defaultValue, options = { maxLength: 100, delay: 50 }) => {
    const initialState = cloneDeep(unwrap(defaultValue));
    const { 
    // Max length saving history states, 100 by default.
    maxLength = 100, 
    // Debounce time for push in milliseconds, 50 by default.
    delay = 50, ...others } = options;
    const history = new History({
        initialState,
        maxLength: maxLength + 1,
        delay,
        ...others
    });
    const store = reactive({
        historyState: unwrap(defaultValue),
        historyLength: 0,
        hasRedo: false,
        hasUndo: false,
        undoCount: 0,
        redoCount: 0
    });
    const isEffective = ref(false);
    const getHistoryLength = () => {
        return (history.$records && history.$records.filter(record => record).length) || 1;
    };
    const getRecord = () => {
        const state = history.get();
        if (state.hasOwnProperty('children') && state.children === undefined)
            delete state.children;
        return state;
    };
    /**
     * Sync state into stateshot
     * @param state
     */
    const snapshot = (state) => {
        history.pushSync(state);
        const historyLength = getHistoryLength();
        store.historyLength = historyLength - 1;
        store.hasUndo = historyLength - 1 > 0;
        store.hasRedo = false;
        store.undoCount = store.historyLength;
        store.redoCount = 0;
    };
    const undo = () => {
        if (store.undoCount === 0 || !store.hasUndo)
            return;
        isEffective.value = true;
        history.undo();
        store.historyState = getRecord();
        store.undoCount -= 1;
        if (store.undoCount === 0)
            store.hasUndo = false;
        store.redoCount += 1;
        store.hasRedo = store.redoCount > 0;
    };
    const redo = () => {
        if (store.redoCount === 0 || !store.hasRedo)
            return;
        isEffective.value = true;
        history.redo();
        store.historyState = getRecord();
        store.redoCount -= 1;
        if (store.redoCount === 0)
            store.hasRedo = false;
        store.undoCount += 1;
        store.hasUndo = store.undoCount > 0;
    };
    const jump = (step) => {
        let jumpStep = step > 0 ? redo : undo;
        for (let i = 0; i < Math.abs(step); i++) {
            jumpStep();
        }
        return getRecord();
    };
    const reset = () => {
        if (store.historyLength <= 0)
            return;
        history.reset();
        // isEffective.value = true
        store.historyState = cloneDeep(initialState);
        store.historyLength = 0;
        store.hasUndo = false;
        store.hasRedo = false;
        store.undoCount = 0;
        store.redoCount = 0;
    };
    watch(() => store.historyState, (newVal) => {
        // undo || redo actions do not need snapshot
        if (isEffective.value) {
            isEffective.value = false;
            return;
        }
        else {
            snapshot(newVal);
        }
    }, {
        // immediate: true,
        deep: true
    });
    return {
        ...toRefs(store),
        undo,
        redo,
        reset,
        jump,
        snapshot
    };
};

/**
 * Reactive window size
 * @param {object} options
 * @return {Ref<number>} WindowSize
 */
const useWindowSize = (options) => {
    const width = ref(isClient ? window.innerWidth : 0);
    const height = ref(isClient ? window.innerHeight : 0);
    const onResize = () => {
        width.value = window.innerWidth;
        height.value = window.innerHeight;
    };
    const { dispatchResize } = useResizeEvent(onResize, options);
    onMounted(dispatchResize);
    return { width, height };
};

export { NOOP, isArray, isBoolean, isClient, isElement, isUndefined, reactiveFn, tryOnMounted, tryOnUnmounted, unwrap, useColor, useCssVars, useDebouncedRef, useElementSize, useEventListener, useFullscreen, useInstance, useMouse, usePalette, usePrecision, useReactiveState, useResizeEvent, useResizeObserver, useShortcut, useStateshot, useWindowSize, wrap };
